<?php

namespace App\Http\Controllers;

use App\Models\Book;
use App\Models\Client;
use App\Models\Dept;
use Illuminate\Http\Request;
use App\Services\ClientService;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Auth;

class ClientController extends Controller
{

    public function __construct(public ClientService $service)
    {
    }
    /**
     * Get the list of clients for an auth() author
     */
    public function index()
    {
        $table_headers = $this->getTableHeaders();

        $allClients = author()->getAllClients();

        // get only first 20 transaction based clients since they're already arranged in latest format
        $allClients = array_slice($allClients, 0, 20, true);

        return view('snit.auth.dashboard.clients.index', compact('allClients', 'table_headers'));
    }

    /**
     * show information about a specific client related to auth() author
     */
    public function show(Client $client)
    {
        // this should have a profile view as opposed to a table view
        // $table_headers  = $this->getTableHeaders();
        $allClients = [];

        $book = Book::find(request()->book);


        if ($book->author->id !== Auth::user()->author->id)
            return redirect()->back(); //authorization(policy)

        foreach ($book->service->getClients() as $orderCreatedAt => $client)
            $allClients[$orderCreatedAt] = $client;

        $allClients = collect($allClients);

        return view('snit.auth.dashboard.clients.show', compact('allClients', 'book'));
    }

    /**
     * display the list of books by auth() author with links to a list of their respective clients.
     */
    public function book_clients_index()
    {
        $allAuthorbooks = author()->books;
        return view(
            'snit.auth.dashboard.clients.clients_by_book.index',
            compact('allAuthorbooks')
        );
    }


    /**
     * display the list of books by auth() author with links to a list of their respective clients.
     */
    public function departmentalClients_index()
    {
        // $deptsInOrg = Dept::where('org_id', author()->org_id)->get();
        $deptsInOrg = authy()->depts;
        // dd(getBooksSoldInDept);
        return view(
            'snit.auth.dashboard.clients.clients_by_dept.index',
            compact('deptsInOrg')
        );
    }


    /**
     * The page that shows a list of all the clients related to a specific book
     * this way we can see all those who have paid or bought a specific book
     */
    public function book_clients_show(Request $request, $bookSlug)
    {
        $table_headers = $this->getTableHeaders();

        // NB:$request['requestedBook'] is used in $this->getClientsBasedOnOrderType($request)
        $requestedBook = $request['requestedBook'] = Book::whereSlug($bookSlug)->first();

        $searchfor = $request->searchfor ?? '';

        $allRelatedClients = $this->getClientsBasedOnOrderType($request);

        return view(
            'snit.auth.dashboard.clients.clients_by_book.show_all_book_clients',
            compact('allRelatedClients', 'table_headers', 'requestedBook', 'searchfor')
        );
    }


    /**
     * The page that shows a list of all the clients related to a specific book that ordered
     * hardcover books only
     * this way we can see all those who have paid or bought a specific book
     */
    public function book_clients_hardcover_show(Request $request, $bookSlug)
    {
        $table_headers = $this->getTableHeaders();

        // use slug to find book
        $requestedBook = $request['requestedBook'] = Book::whereSlug($bookSlug)->first();

        $searchfor = $request->searchfor ?? '';

        $allRelatedClients = $this->getClientsBasedOnOrderType($request, 'hardcover');

        return view(
            'snit.auth.dashboard.clients.clients_by_book.hardcover_clients',
            compact('allRelatedClients', 'table_headers', 'requestedBook', 'searchfor')
        );
    }


    /**
     * The page that shows a list of all the clients related to a specific book that ordered
     * hardcover books only
     * this way we can see all those who have paid or bought a specific book
     *
     */
    public function book_clients_ebook_show(Request $request, $bookSlug)
    {
        $table_headers = $this->getTableHeaders();
        $requestedBook = $request['requestedBook'] = Book::whereSlug($bookSlug)->first();
        $searchfor = $request->searchfor ?? '';

        $allRelatedClients = $this->getClientsBasedOnOrderType($request, 'ebook');
        // dd('fisi');
        return view(
            'snit.auth.dashboard.clients.clients_by_book.ebook_clients',
            compact('allRelatedClients', 'table_headers', 'requestedBook', 'searchfor')
        );
    }



    // helpers


    /**
     * gets clients for a specific book based on (order)type i.e ebook, hardcover or both
     *
     * @param Request $request
     * @param string $bookSlug
     * @param string|null $type refers to the order type: ebook or hardcover. 
     * 
     * if $type is not provided, then both ebook and hardcover order clients are returned.
     * 
     * @return \Illuminate\Support\Collection
     */
    public function getClientsBasedOnOrderType(Request $request, string $type = null): Collection
    {
        $allClients = [];

        // use slug to find book
        $requestedBook = $request['requestedBook'];

        // dd($requestedBook->service->getClients($type), $request->all(), $request->bookSlug);
        // we ain't using unique clients here, wanna see it all.
        foreach ($requestedBook->service->getClients($type) as $key => $client)
            $allClients[$key] = $client;

        return $allRelatedClients = $this->filteredResults($request, collect($allClients)) ?? collect($allClients);

    }

    public function getTableHeaders()
    {
        return $table_headers = [
            'ORG',
            'Client Name',
            'Mat. Number',
            // "isbn",
            'Client Phone',
            'Purchase <br> Date',
            'Book Purchased',
            'Purchase <br> Type',
            'income(₦)',
            // 'ebook <br>income(₦)',
        ];
    }

    /**
     * This function filters the client results for a specific book, depending on whether the user(author)
     * (requested or ) searched for a mat_num or phone_number etc.
     * 
     * It finally returns all clients that have a property value that is 85% or 75% match with the searchfor
     * eg. it searches for the mat_num property (column) on the client and compares it with the 
     * mat_num from the request->searchfor
     *
     * @param [type] $request
     * @param [type] $allRelatedClients
     * @return void
     */
    public function filteredResults($request, $allRelatedClients)
    {
        if ($request->searchfor && $request->filterby) {

            $searchfor = $request->searchfor;
            /**
             * @var string $filter the model property being compared in the db vs the request searchfor
             */
            $filter = $request->filterby;

            // return only matches that are at least above 35% accurate

            // if($filter == 'book_name'){
            //     return $allRelatedClients = $allRelatedClients->filter(function ($client) use ($filter, $searchfor) {
            //         $result = similar_text($client->itemPurchased->$filter, $searchfor, $perc);
            //         return ($perc > 85) ?: ($perc >= 75);
            //     });
            // }
            return $allRelatedClients = $allRelatedClients->filter(function ($client) use ($filter, $searchfor) {
                $result = similar_text($client->client_model->$filter, $searchfor, $perc);
                return ($perc > 85) ?: ($perc >= 75);
            });
        }
    }
}
