<?php

namespace App\Models;

use Illuminate\Support\Collection;
use Illuminate\Database\Eloquent\Model;
use App\PaymentGatewayServices\Flutterwave;
use Illuminate\Database\Eloquent\Factories\HasFactory;

class Developer
{
    use HasFactory;

    /**
     * @const float FEE_IN_PERC represents developer fee in percentage
     */
    public const TOTAL_SERVICE_FEE_IN_PERC = 0.055;
    /**
     * RATE FOR FLUTTERWAVE IS 4.1% 
     * RATE FOR FLUTTERWAVE [4.1% + 1.4% = 5.5%]
     */
    public const FEE_IN_PERC_FLW = 0.041;
    /**
     * RATE FOR OTHER PGS IS 4% 
     * RATE FOR FLUTTERWAVE [4% + 1.5% = 5.5%]
     */
    public const FEE_IN_PERC_GEN = 0.04;
    // this is a model for developer related queries. it has no linked tables to it

    /**
     * Get total number of books published on the platform
     */
    public function getAllBooksPublished()
    {
        return Book::all()->count();
    }

    /**
     * returns the total number of books sold from each order.
     *  In a nutshell, we are getting every item that
     * has been added to cart and made it to the order basket. 
     * This is for every order registered on the platform.
     */
    public function countOfBookSold()
    {
        $sumOfBookSold = 0;
        foreach (Author::all() as $author) {
            $sumOfBookSold += $author->totalBookSold();
        }

        return $sumOfBookSold;
    }


    /**
     * This represents all the sales generated by the app for authors. excluding service charges and tax
     * The value returned is the sum of all the sales or payout for each authors.
     */
    public function getTotalSalesGeneratedForAuthors(): float
    {
        $totalSalesGenerated = 0.00;

        foreach (Author::all() as $author) {
            $totalSalesGenerated += $author->totalSalesFromBooks();
        }
        // dd($totalSalesGenerated);
        return $totalSalesGenerated;
    }

    /**
     * Get the total amount of profit generated across all authors
     */
    public function getTotalProfitGeneratedForAuthors()
    {
        $totalProfitGenerated = 0.00;

        foreach (Author::all() as $author) {
            $totalProfitGenerated += $author->presentTotalProfit();
        }

        return $totalProfitGenerated;
    }

    /**
     * Get the total number of authors registered on the platform
     */
    public function getRegisteredAuthors()
    {
        return Author::all()->count();
    }

    /**
     * This computes the profits of all the stakeholders in the app viz; author, pgs, dev
     *
     * @param  string  $who refers to the person or company whose (net)profit is being requested
     * @param  string  $duration specifies a timeframe/time interval for use in querying the db
     */
    public function getTotalStakeholderIncome($who, $duration = null)
    {
        /**
         * programming tag assigned to each stakeholder income for every order
         */
        (string) $prop = '';
        $totalProfitGenerated = 0.00;

        try {
            match ($who) {
                'authors', 'author' => $prop = 'authors_remit',
                'pgs' => $prop = 'pgs_remit',
                'dev', 'developer' => $prop = 'dev_remit',
                default => throw new \Exception("Error Processing Request : Unknwon prop: '$who' ", 1)
            };

            if ($duration) {
                match ($duration) {
                    'this_week' => $duration = [now()->startOfWeek(), now()->endOfWeek()],
                    'this_month' => $duration = [now()->startOfMonth(), now()->endOfMonth()],
                    'this_year' => $duration = [now()->startOfYear(), now()->endOfYear()],
                    'all', '*', '', null => $duration = [],
                    default => throw new \Exception("Error Processing Request : Unknown duration : '$duration' is ", 1)
                };
            }
        } catch (\Throwable $th) {
            dd($th->getMessage());
        }

        /**
         * @var Collection $requestedOrders refers to all user requested orders within given timeframe
         */
        $requestedOrders = ($duration)
            ? Order::successful()->whereBetween('created_at', $duration)->get()
            : Order::successful()->get();

        // dd($requestedOrders);
        foreach ($requestedOrders as $order)
            $totalProfitGenerated += $order->$prop; // add code below 4 additional pgs fees.

        return $totalProfitGenerated;
    }


    public function getTotalProfitGeneratedForPGS()
    {
        (float) $total_PGS_fees = 0.00;

        foreach (Order::all() as $order)
            $total_PGS_fees += $order->pgs_remit; //normal 1.4%

        return $total_PGS_fees;
    }

    /**
     * This function is for whenever pgs decides to add additional fees
     * to their service.
     *
     * @param Order $order
     * @return float
     */
    public function calculateExtraFee4PGS(): bool|float
    {
        if (Flutterwave::hasAddedFees())
            return false;

        $allOrdersAbove2400 = Order::above2400()->successful()->get();
        (int) $sumOfOrdersAbove2400 = $allOrdersAbove2400->count();

        // dd($allOrdersAbove2400);
        foreach ($allOrdersAbove2400 as $order)
            echo $order->total_amount . '<br>';

        return $totalCost = $sumOfOrdersAbove2400 * Flutterwave::getAdditionalFee();
    }
}
