<?php

namespace App\PaymentGatewayServices;

use stdClass;
use App\Models\Order;
use PharIo\Manifest\Url;
use Illuminate\Support\Arr;
use App\Contracts\PaymentGateWay;
use Illuminate\Support\Facades\DB;
use Illuminate\Http\Client\Response;
use Illuminate\Support\Facades\Http;
use App\Traits\AppTraits\PaymentGatewayTrait;

class Flutterwave implements PaymentGateWay
{
    use PaymentGatewayTrait;

    /**
     * property that stores the information whether FLW has added fees
     */
    public static bool $hasAddedFees = false;

    /**
     * obtain the value of flw additional fees eg. 100, 200 etc
     *
     * @var float $additionalFees
     */
    public static float $additionalFees;


    public function __construct()
    {

    }



    /**
     * get Headers for Flw API Request
     */
    public function getHeaders(): array
    {
        return $headersArray = [
            'Authorization' => 'Bearer ' . env('FLW_SECRET_KEY'),
            "Content-Type" => "application/json",
        ];
    }

    /**
     * obtain the value of flw main fees as percentage eg. 0.014, 0.20 etc
     */
    public static function getRate()
    {
        return 0.014;
    }

    /**
     * Checks if Flutterwave has additional charges
     *
     * @return boolean
     */
    public static function hasAddedFees(): bool
    {
        return (bool) self::getAdditionalFee();
    }

    /**
     * returns the additional charges for Flutterwave
     *
     * @return float
     */
    public static function getAdditionalFee(): float
    {
        return self::$additionalFees;
    }

    /**
     * Prepare FLW data for transacting with the FLW API
     * don't touch
     */
    public function buildTrxData(): null|array
    {
        $params = session('paymentParams'); //important

        // destructure the needed variables
        ['cartTotal' => $cartTotal, 'client' => $client] = $params;

        $client->only(['fname', 'email', 'phone_num', 'mat_num']); //customer must be an object

        $clientData = [
            'name' => $client->getFullName(),
            'email' => $client->email,
            'phone_number' => $client->mat_num,
        ];

        $trxRef = generateTxRef();

        $reqBody = [
            // compulsory fields
            'tx_ref' => $trxRef,
            'amount' => $cartTotal,
            'currency' => 'NGN',
            'redirect_url' => route('flw.confirm', ['mytx_ref' => $trxRef]),

            // =========optional fields============
            // send customer info to flw
            'customer' => $clientData,
            'customizations' => $this->getDefaultCustomizations(), //customizations must be an object
        ];

        $validateCartTotal = (float) $reqBody['amount'] === (float) $cartTotal;

        return (!$validateCartTotal) ? null : $reqBody;
    }


    public function getDefaultCustomizations(): stdClass
    {
        // customize flw payment appearance
        $customizations = new stdClass();
        $customizations->title = 'AIMS Book Purchase Payments';
        $customizations->logo = 'https://www.fintechfutures.com/files/2021/11/Flutterwave-logo.png';

        return $customizations;
    }


    /**
     * Make a FLW payment request, via the FLW API.
     * This is our app making a connection to flw 
     * and sending the required data for payment initiation.
     * Flw then makes the payment on our behalf if all is well.
     */
    public function createPayment(array $flwTrxData): Response
    {
        $flwPaymentUrl = 'https://api.flutterwave.com/v3/payments';

        return $res = Http::withHeaders($this->getHeaders())->post($flwPaymentUrl, $flwTrxData);
    }

    /**
     * Make a FLW payment request via the FLW API
     * This is our app authenticating and validating with flw successfully
     * the result is a flw payment link that signifies that we can now make payment via flw
     */
    public function prePaymentValidations(Response $response): string|null
    {
        // Defensive programming here !don't touch
        // boolean expressions declaration
        $res = json_decode($response->body(), 1);
        $validateFLWstatus = ($res['status'] == 'success');
        $paymentLink = $res['data']['link'];

        // grouping the validations
        return ($validateFLWstatus && $paymentLink) ? $paymentLink : null;
    }


    /**
     * Make a FLW payment verification request, via the FLW API.
     * Using the flw generated transaction_id
     * @param string $flwTrxId represents the transaction_id generated by flw after payment was made. 
     */
    public function verifyPaymentRequest($flwTrxId): Response
    {
        $flwPostPaymentConfirmUrl = "https://api.flutterwave.com/v3/transactions/$flwTrxId/verify";

        return $res = Http::withHeaders($this->getHeaders())->get($flwPostPaymentConfirmUrl);
    }


    /**
     * Proper FLW verification implementation (post payment)
     */
    public function postPaymentConfirmation()
    {

        $relatedOrder = Order::where('tx_ref', request()->mytx_ref)->first();
        $flwTrxId = request()->transaction_id;

        session()->put('paymentParams.flwTrxId', $flwTrxId); //for future purposes

        $paymentSuccessful = request()->status == 'successful';

        if (!($paymentSuccessful and $flwTrxId and $relatedOrder->id)) {
            // truthyExpectations are optional
            $truthyExpectations = ['paymentStatus' => $paymentSuccessful, 'flwTrxId' => $flwTrxId, 'order' => $relatedOrder->id];
            return $this->handleErrorPayments($relatedOrder, $truthyExpectations);
        }
        $validatedTrx = $this->postPaymentValidation($relatedOrder, $flwTrxId);

        if (!$validatedTrx) {
            return view(
                'snit.thankyou.error',
                [
                    'order_id' => $relatedOrder->identifier,
                    'status' => request()->status,
                    'reason' => request()->status,//optional
                ]
            );
        }

        return $this->generateViewForSuccessfulPayment($relatedOrder);
    }


    private function handleErrorPayments($relatedOrder, $truthyExpectations)
    {
        $reason = array();
        foreach ($truthyExpectations as $key => $expectation)
            $reason[$key] = $expectation ? null : 'error processing ' . $key;
        // update request status for this order
        $relatedOrder->update(['status' => request()->status]);

        return view(
            'snit.thankyou.error',
            [
                'order_id' => $relatedOrder->identifier,
                'status' => request()->status,
                'reason' => $reason,//optional
            ]
        );
    }
    /**
     * this is where you do post validation before giving value or changing
     * the state of your app
     */
    private function postPaymentValidation(Order $relatedOrder, $flwTrxId)
    {
        // my expectations
        $myExpectedAppFee = (float) strval($relatedOrder->total_amount * (1.4 / 100));//fake 1.4%

        $myExpectedAmountToBeSettled = (float) strval($relatedOrder->total_amount - $myExpectedAppFee);

        // possible refactor needed here. This NGN is hardcoded, might cause issues for other nations
        $myExpectedCurrency = 'NGN';

        $expectedTxRef = $relatedOrder->tx_ref;
        // ======================================================================

        $finalResponse = $this->verifyPaymentRequest($flwTrxId);
        $resBody = json_decode($finalResponse->body(), 1);

        $statusCode200 = $finalResponse->getStatusCode() == 200;// check1
        $validResponse = $resBody['status'] == 'success';// check2

        $mytx_ref = strval(Arr::get($resBody, 'data.tx_ref'));
        $validTxRef = $mytx_ref === $expectedTxRef;// check3

        // money matters
        $chargedAmount = (float) strval(Arr::get($resBody, 'data.charged_amount'));
        $amount = (float) strval(Arr::get($resBody, 'data.amount'));
        $validChargedAmount = ($amount === $chargedAmount) === $relatedOrder->total_amount; // check4

        $appFee = (float) strval(Arr::get($resBody, 'data.app_fee'));
        $verifiedAppFee = $myExpectedAppFee === $appFee; // check5

        $amountSettled = (float) strval(Arr::get($resBody, 'data.amount_settled'));
        $validAmountSettled = $myExpectedAmountToBeSettled === $amountSettled; // check6

        $currency = strval(Arr::get($resBody, 'data.currency'));
        $validCurrency = $currency === $myExpectedCurrency; // check7

        // grouping the checks
        $validTrxData = $statusCode200 && $validResponse && $validTxRef;
        $validFinancialData = $validChargedAmount && $verifiedAppFee && $validAmountSettled && $validCurrency;
        $validTrxAndFinancialData = $validTrxData && $validFinancialData; // strict comparisons

        return (env('APP_ENV') !== 'local') ? $validTrxAndFinancialData : $validTrxData;
    }


}
