<?php

namespace App\PaymentGatewayServices;

use App\Models\Order;
use PharIo\Manifest\Url;
use App\Contracts\PaymentGateWay;
use Illuminate\Http\Client\Response;
use Illuminate\Support\Facades\Http;
use Illuminate\Support\Facades\Mail;
use App\Traits\AppTraits\PaymentGatewayTrait;
use App\Mail\OrderVerificationFailedWithPGSMail;

class Rexpay implements PaymentGateWay
{

    use PaymentGatewayTrait;
    /**
     * property that stores the information whether Rexpay has added fees
     */
    public static bool $hasAddedFees = false;

    /**
     * obtain the value of Rexpay additional fees eg. 100, 200 etc
     *
     * @var float $additionalFees
     */
    public static float $additionalFees = 0.00;

    /**
     * obtain the value of Rexpay main fees as percentage eg. 0.014, 0.20 etc
     *
     * @var float $additionalFees
     */
    public static float $rate = 0.014;


    public function init()
    {
        dd('init rexpay');
    }

    /**
     * Checks if Rexpay has additional charges
     *
     * @return boolean
     */
    public static function hasAddedFees(): bool
    {
        return self::$hasAddedFees;
    }

    /**
     * returns the additional charges for Rexpay
     *
     * @return float
     */


    /**
     * get Headers for Rexpay API Request
     */
    public function getHeaders(): array
    {
        return $headersArray = [
            // 'Authorization' => 'Basic YWNjZXNzdHJ1c3RiYW5rMDlnbWFpbGNvbTphZDE1YjZiNTE4ZjM5YzM1ZmZmZDUzMjA5MTg0ZjNiMDUzYjJjMzYxNjQ0MGMwZTM5YTM0M2M2ODJkMGU=',
            'Authorization' => env('LIVE_REXPAY_SECRET_KEY_FOR_BASIC_API_AUTH'),
            "Content-Type" => "application/json",
        ];
    }
    public static function getAdditionalFee(): float
    {
        return self::$additionalFees;
    }

    /**
     * Prepare Rexpay data for transacting with the Rexpay API
     * don't touch
     */
    public function buildTrxData(): null|array
    {
        $params = session('paymentParams'); //important

        // destructure variables needed
        ['cartTotal' => $cartTotal, 'client' => $client] = $params;

        $trxRef = generateTxRef();

        // the request body : rexpayTrxDataAsArray
        $reqBody = [
            // compulsory fields
            'reference' => $trxRef,
            'amount' => $cartTotal,
            'currency' => 'NGN',
            "userId" => 'accesstrustbank09@gmail.com',
            "callbackUrl" => route('rexpay-redirect-url', ['mytx_ref' => $trxRef]),

            // =========optional fields============
            // send customer info to Rexpay
            'metadata' => [
                'description' => 'AIMS Book Purchase Payments',
                'matric_num' => $client->mat_num,
                'customer' => $client->fname ?? 'Peace Onyeka',
            ],
        ];
        $validateCartTotal = (float) $reqBody['amount'] === (float) $cartTotal;

        return (!$validateCartTotal) ? null : $reqBody;
    }



    /**
     * Make a Rexpay payment request via the Rexpay API
     * This is our app making a connection to Rexpay 
     * and sending the required data for payment 
     */
    public function createPayment(array $reqBody): Response
    {
        $base = env('REXPAY_LIVE_API_BASE_URL');
        /**
         * @var string $createPaymentLink this is the request url to visit, when you want 
         * to ask for a payment link to be generated from Rexpay
         */
        $createPaymentLink = $base . "/api/pgs/payment/v2/createPayment";

        return $res = Http::withHeaders($this->getHeaders())->post($createPaymentLink, $reqBody);
    }


    /**
     * Make a Rexpay post payment verification request via the Rexpay API
     * this was the older version, it has inconsistencies
     */
    public function verifyPaymentRequestV2(Order $order): Response
    {
        $base = env('REXPAY_LIVE_API_BASE_URL');
        /**
         * @var string $rexpayPostPaymentVerificationUrl is the request url to visit, when you want 
         * to ask for a payment link to be generated from Rexpay
         */
        $rexpayPostPaymentVerificationUrl = "$base/api/cps/v1/getTransactionStatus";
        // /api/pgs/payment/v1/getPaymentDetails/{reference}

        return $res = Http::withHeaders($this->getHeaders())
            ->post($rexpayPostPaymentVerificationUrl, ['transactionReference' => $order->tx_ref]);
    }


    /**
     * Make a Rexpay post payment verification request via the Rexpay API
     */
    public function verifyPaymentRequest(Order $order): Response
    {
        $base = env('REXPAY_LIVE_API_BASE_URL');

        $trxRef = $order->tx_ref;

        /**
         * @var string $paymentVerificationUrl is the request url to visit, when you want 
         * to ask for a payment link to be generated from Rexpay
         */
        $paymentVerificationUrl = "$base/api/pgs/payment/v1/getPaymentDetails/$trxRef";

        return $res = Http::withHeaders($this->getHeaders())->get($paymentVerificationUrl);
    }


    /**
     * Make a Rexpay payment request via the Rexpay API
     * This is our app making a connection to Rexpay 
     * and sending the required data for payment 
     */
    public function authenticateRexpayAPI(array $rexpayTrxData): Response
    {
        $rexpayUserId = env('LIVE_REXPAY_USERNAME');
        $rexpayUserEmail = env('LIVE_REXPAY_USER_EMAIL');
        $authRoute = "/api/pgs/clients/v1/get/$rexpayUserId/token?email=$rexpayUserEmail";

        return $response = Http::withHeaders([
            'Authorization' => 'Basic ' . env('LIVE_REXPAY_SECRET_KEY'), // 'X-Second' => 'bar'
        ])->post(env('REXPAY_LIVE_API_BASE_URL') . $authRoute);
    }

    /**
     * Make a Rexpay payment request via the Rexpay API
     * This is our app authenticating and validating with Rexpay successfully
     * the result is a Rexpay payment link that signifies that we can now make payment via Rexpay
     */
    public function prePaymentValidations(Response $response): string|null
    {
        // You must manually get the response body (ain't like flw)
        $resBody = json_decode($response->body(), 1); //they send body in json by default
        $passedTests = $compulsory = $this->validateCompulsoryRules($resBody);

        // $optional = $this->validateOptionalRules($resBody);

        return $paymentUrl = ($passedTests) ? $resBody['paymentUrl'] : null;

    }



    function tipsToHelpInRexpayValidation(): void
    {
        // once a payment fails, you can't reuse that payment link again, even if you made a mistake, 
        // you'll have to redo the whole process
        //1. status must be CREATED
        // 'ONGOING' OR 'SUCCESS' MEANS THE TRX_ID HAVE ALREADY BEEN UTILIZED
        //2. "paymentUrl": "https://checkout-dev.globalaccelerex.com/pay/17079048CrEub5AzRQ",

        // 3.paymentUrlReference and externalPaymentReference are always identical. see below
        // "paymentUrlReference": "17079048CrEub5AzRQ",
        // "externalPaymentReference": "17079048CrEub5AzRQ",

        // 4."fees": 3.00, //always validate fees

        // skip: "paymentChannel": "DEFAULT" or "CARD",//not necessary, they can pay with bank tt.
        // 5. validate "clientId": "accesstrustbank09gmailcom",

        // LATER WE SHALL DO DESTRUCTURING
        // $paymentLink = $resBody['paymentUrl'];
    }


    private function validateCompulsoryRules(array $resBody): bool
    {
        // dd($resBody);
        // validation 1: status must be 'CREATED'
        $validStatus = ($resBody['status'] == 'CREATED');

        // validation 2: fees must be accurate
        // rexpay values are json stringified, so pls convert to string before doing any further comparisons
        $cartTotal = session('paymentParams')['cartTotal'];

        // fees should be the same wehen rounded to 2dp and the difference in cost should never be more than N1

        $validFees = $this->validFees($resBody, $cartTotal);

        //check that they're not scamming me

        // validation3 : auth check [make dem no divert my funds]
        $authorizedUser = $resBody['clientId'] == env('LIVE_REXPAY_USERNAME');
        // grouping the validations
        return $compulsoryRules = $validStatus && $validFees && $authorizedUser;

    }

    public function validFees($resBody, $cartTotal): bool
    {
        // fees should be the same wehen rounded to 2dp and the diff <= N1
        $rexpayFees = (strval($resBody['fees']));
        $agreedFees = round(strval($cartTotal * 0.015), 2); //15% total rexpay fee per trx
        $validFees = $rexpayFees == $agreedFees;  //check that they're not scamming me

        $reasonableDiff = (abs($agreedFees - $rexpayFees) <= 0.5);
        return $validFees && $reasonableDiff;
    }
    private function validateOptionalRules(array $resBody): bool
    {
        // validation 4: identical references
        // pls ask if this is always supposed to be the same before adding it to your must haves
        $identicalRefs = $resBody['paymentUrlReference'] == $resBody['externalPaymentReference'];

        // validation 5: validatePaymentUrl
        $endOfPaymentUrl = endOfUrl($resBody['paymentUrl']);
        $validPaymentUrl = $endOfPaymentUrl == $resBody['paymentUrlReference'];

        // grouping the validations
        return $optionalRules = $identicalRefs && $validPaymentUrl;
    }


    private function orderNotFoundResponseAsView()
    {
        return view(
            'snit.thankyou.error',
            [
                'order_id' => 'not found',
                'reason' => 'Transaction Failed',
                'status' => request()->status,
            ]
        );
    }
    /**
     * Proper Rexpay verification implementation after payment API endpoint 
     * has been called. Post payment validation happens here
     */
    public function rexpayPostPaymentConfirmation()
    {
        $relatedOrder = Order::where('tx_ref', request()->mytx_ref)->first();
        if (!$relatedOrder)
            return $this->orderNotFoundResponseAsView();

        // make api call to endpoint to validate trx data and financial data
        $isTrxValid = $this->postPaymentValidation($relatedOrder);

        $validatedTrx = is_bool($isTrxValid) ? $isTrxValid : null;

        $problemWithTrx = is_array($isTrxValid) ? $isTrxValid : null;

        if ($problemWithTrx) {
            if (strstr($problemWithTrx['reason'], 'invalid payment response')) {

                // at this point we know that the trx did not go through, so we rename
                // the variable to reflect that in our code
                $failedTrx = $problemWithTrx;

                // update status with what was sent from rexpay api
                $relatedOrder->update(['status' => $failedTrx['status']]);

                // this means rexpay didn't recieve payment (eg. insufficient funds)
                // hence no need to send developer any email. just redirect to error page
                return view(
                    'snit.thankyou.error',
                    [
                        'order_id' => $relatedOrder->identifier,
                        'status' => $failedTrx['status'],
                        'reason' => $failedTrx['reason'],
                    ]
                );
            }

            //mail developer: successful trx but could not validate infor with rexpay
            // ==> the problem with the trx most times lies in the fees (inconsistencies)
            Mail::to(env('DEV_EMAIL1'))->send(new OrderVerificationFailedWithPGSMail($relatedOrder));
        }
        return $this->generateViewForSuccessfulPayment($relatedOrder);


    }






    /**
     * this is where you do post validation before giving value or changing
     * the state of your app. In this method we call a special method that queries
     * the rexpay API for the status of our transaction.
     * it returns a boolean 'true' if successful and an array of errors if failed
     */
    private function postPaymentValidation(Order $relatedOrder): array|bool
    {

        // query transaction info from rexpay 
        $rexpayResponseForTrxQuery = $this->verifyPaymentRequest($relatedOrder);
        $resBody = json_decode($rexpayResponseForTrxQuery->body(), 1);

        // ==================================STEP 1=========================================
        $statusCode200 = $rexpayResponseForTrxQuery->getStatusCode() === 200;// check1 group
        $validResponseDescription = $resBody['status'] === "SUCCESS";// check1 group

        $validResponse = $statusCode200 && $validResponseDescription;// check1
        // dd($resBody);
        if (!$validResponse) {
            return $this->getErrorResponse($resBody, [
                'reason' => 'invalid payment response : ' . $resBody['providerResponse'],
            ]);
        }


        // ==================================STEP 2=========================================
        // my expectations
        $orderTotalAmount = (float) strval($relatedOrder->total_amount);

        $myExpectedAppFee = (float) strval($orderTotalAmount * (1.5 / 100)); //1.5%

        $myExpectedCurrency = 'NGN';

        $expectedTxRef = $relatedOrder->tx_ref;
        // ======================================================================


        // ============================================================================================
        // ==================================STEP 3: VARIOUS API RESULT CHECKS=========================
        // ============================================================================================

        // ==================================CHECK 1 : PAYMENT REFERENCE=========================
        $paymentReference = strval($resBody['referenceId']);
        $validTxRef = $paymentReference === $expectedTxRef;// check2

        if (!$validTxRef) {
            return $this->getErrorResponse($resBody, [
                'reason' => 'invalid payment reference : ' . $paymentReference
            ]);
        }


        // ==================================CHECK 2 : CHARGED AMOUNT=========================
        // money matters --done
        $chargedAmount = (float) strval($resBody['amount']);
        $validChargedAmount = $chargedAmount === $orderTotalAmount; // check3
        if (!$validChargedAmount) {
            return $this->getErrorResponse($resBody, [
                'reason' => 'invalid charged Amount : ' . $chargedAmount
            ]);
        }


        // ==================================CHECK 3 : APP FEE=========================
        $appFee = (float) strval($resBody['fees']);
        $verifiedAppFee = $myExpectedAppFee === $appFee; // check4
        if (!$verifiedAppFee) {
            return $this->getErrorResponse($resBody, [
                'reason' => 'invalid app fee : ' . $appFee
            ]);
        }


        // ==================================CHECK 4 : CURRENCY=========================
        $currency = strval($resBody['currency']);
        $validCurrency = $currency === $myExpectedCurrency; // check5

        if (!$validCurrency) {
            // mail developer about app fee discrepancy
            // Mail::to(env('DEV_EMAIL1')->email)->send(new PGSAppCurrencyDiscrepancyMail($relatedOrder));

            return $this->getErrorResponse($resBody, [
                'reason' => 'invalid currency : ' . $currency
            ]);
        }

        return true;
        // return ($validResponse && $validTxRef && $validChargedAmount && $verifiedAppFee && $validCurrency);
    }


    private function getErrorResponse($resBody, array $errorBag)
    {
        $defaultBody = [
            'status' => $resBody['status'],
        ];
        return $errorBag + $defaultBody;

    }
}