<?php

namespace App\Services\Book;

use stdClass;
use App\Models\Book;
use App\Models\Order;
use App\Models\Client;
use App\Models\Developer;
use App\Models\Publisher;
use Illuminate\Support\Carbon;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Auth;
use Illuminate\Http\RedirectResponse;
use App\Http\Requests\StoreBookRequest;

class BookService
{


    // Refactor operation
    /**
     * general book methods
     * ebook methods
     *  hardcover book methods
     * pricing and sales book methods (ebook and hardcover)
     */


    public function __construct(public Book $model)
    {
    }

    /**
     * returns an assoc Array where the key for every client is the 
     * order->date from the related Order
     * @param string $type refers to the type of the book ordered, ebook or hardcover.
     * 
     * $type = 'ebook' 
     * 
     * or
     *  
     * $type = 'hardcover'
     * 
     * please note that the $type argument only accepts an enum of ['ebook', 'hardcover']
     * nothing else is permitted 
     * @return array
     */
    public function getClients(string $type = null): array
    {
        $subOrders = $clients = [];
        $completedOrders = Order::successful()->get();

        foreach ($completedOrders as $order) {
            // check if book is represented in the order
            if ($foundCartItem = $this->findItemInOrder($order)) {
                if ($type and $foundCartItem->options->type !== $type) {
                    continue;
                }



                $foundCartItemObj = new stdClass;

                // client information
                $foundCartItemObj->client_id = $order->client_id;
                $foundCartItemObj->client_model = Client::find($order->client_id);

                // book model or item purchased information
                $foundCartItemObj->itemPurchased = $foundCartItem;

                // make the keys also accessible on the $foundCartItemObj
                $foundCartItemObj->created_at_key = "$order->created_at";

                // store orders [indexed] by the date of creation
                // why? this will help us sort orders by date using their index|keys
                $subOrders["{$order->created_at}"] = $foundCartItemObj;
            }
        }

        ksort($subOrders); //  sort client ids according to the latest (using their keys)

        // get associated client models for each client ID
        foreach ($subOrders as $key => $subOrder)
            $clients[$key] = $subOrder;

        return $clients; //return all clients after looping
    }


    public function totalHardcoverOrders($duration = null)
    {
        return $this->getCollectionOfTotalBookSold($duration)?->where('options.type', '==', 'hardcover');
    }


    public function totalEbookOrders($duration = null)
    {
        return $this->getCollectionOfTotalBookSold($duration)?->where('options.type', '==', 'ebook');
    }

    public function presentTotalEbookSales($duration = null)
    {
        return $this->totalEbookOrders($duration)->sum('options.original_price');
    }

    /**
     * Gets and presents the total hardcover sales for a specific book
     * presenter is for authors
     *
     * @param [type] $duration
     * @return float
     */
    public function presentTotalHardcoverSales($duration = null): float
    {
        return $this->totalHardcoverOrders($duration)->sum('options.original_price');
    }


    /**
     * Gets and presents the total hardcover sales PROFIT  for a specific book
     * presenter is for authors
     *
     * @param [type] $duration
     * @return float
     */
    public function presentTotalHardcoverSalesProfit($duration = null): float
    {
        return $this->presentTotalHardcoverSales() * profitMargin('hardcover');
    }



    /**
     * Gets and presents the total hardcover sales PROFIT  for a specific book
     * presenter is for authors
     *
     * @param [type] $duration
     * @return float
     */
    public function presentTotalEbookSalesProfit($duration = null): float
    {
        return $this->presentTotalEbookSales() * profitMargin('ebook');
    }


    //  developer
    public function getTotalHardcoverSales($duration = null)
    {
        return $this->totalHardcoverOrders($duration)->sum('price');
    }

    public function getTotalEbookSales($duration = null)
    {
        return $this->totalEbookOrders($duration)?->sum('price');
    }
    //dev eol

    //  authors

    public function getPublisher($attr)
    {
        return $attr ? Publisher::find($this->model->publisher_id)?->$attr : Publisher::find($this->model->publisher_id);
    }


    protected function findItemInOrder(Order $order)
    {
        $order_content = collect(json_decode($order->content)); //json -> (obj) -> collection for easy traversal with where()

        return $found = $order_content->where('id', $this->model->id)->first() ?? false; // return cartItem related to specific book
    }

    /**
     * checks if  a book already exists in the db
     *
     * @param StoreBookRequest $request
     * @return void
     */
    public function bookExists(StoreBookRequest $request)
    {
        return Auth::user()->author->books()->whereName($request->name)->where('ISBN', $request['ISBN'])->first();
    }


    public function createBook(StoreBookRequest $request)
    {
        return Auth::user()->author->books()->create($request->validated());
    }

    /**
     * update a book in db
     *
     * @param StoreBookRequest $request
     * @param Book $book
     * @return void
     */
    public function updateBook(StoreBookRequest $request, Book $book): bool
    {
        return $book->update($request->validated());
    }


    public function deleteBook(Book $book)
    {

        return ($book->delete()) ? back()->with('success', 'Book Deleted') : null;
    }
    public function redirectToCreateNewDocumentForBook(Book $book): RedirectResponse
    {
        return redirect()->route('docs.create', ['uploadfiles' => true, 'book' => $book])
            ->with('success', 'Book Updated successfully');
    }

    public function redirectToEditDocumentForBook(Book $book): RedirectResponse
    {
        return redirect()->route('docs.edit', ['uploadfiles' => true, 'book' => $book, 'doc' => $book->doc])
            ->with('success', 'Book Updated successfully');
    }

    public function redirectWithBookFailedToUpdateMessage()
    {
        return back()->with('error', 'Oops! Try again... Book Updated Failed!');
    }


    /**
     * returns orders between a specified duration or timeframe. Taking two variables of: start & end duration
     */
    private function getOrdersBetween($start, $end): Collection
    {
        // here we are only looking at orders within a certain timeframe..
        $totalFound = [];

        $orderInTimeframe = Order::successful()
            ->whereBetween('created_at', [$start, $end])
            ->get();

        foreach ($orderInTimeframe as $order)
            $totalFound[] = $this->findItemInOrder($order); //store collections in array for easy filtration

        return collect($totalFound)->filter(fn($elem) => !empty ($elem)); //filter out empty or null values
    }



    /**
     * returns the quantity of this book (model) sold based within a timeframe or duration
     *
     * @param $duration refers to the time interval.
     */
    public function getTotalNumberSold($duration = null): int
    {
        return match ($duration) {
            'last_week' => $this->getCollectionOfTotalBookSoldLastWeek()->count(),
            'this_week' => $this->getCollectionOfTotalBookSoldThisWeek()->count(),

            'last_month' => $this->getCollectionOfTotalBookSoldLastMonth()->count(),
            'this_month' => $this->getCollectionOfTotalBookSoldThisMonth()->count(),

            'last_year' => $this->getCollectionOfTotalBookSoldLastYear()->count(),
            'this_year' => $this->getCollectionOfTotalBookSoldThisYear()->count(),

            'today' => $this->getCollectionOfTotalBookSoldToday()->count(),

            //   past data
            'yesterday' => $this->getCollectionOfTotalBookSoldYesterday()->count(),

            default => $this->getCollectionOfTotalBookSold()->count()
        };
    }

    /**
     * returns all books sold (from the orders table) in a duration, referenced (keyed) by their respective order_ids
     *
     * returns $totalFound[$order->id] = $this->findItemInOrder($order)
     * @param string $duration
     * @return Collection
     */
    public function getCollectionOfTotalBookSold($duration = null): Collection
    {
        $totalFound = [];
        if ($duration) {
            try {
                match ($duration) {
                    'today' => $duration = [now()->startOfDay(), now()->endOfDay()],
                    'yesterday' => $duration = [now()->subDay()->startOfDay(), now()->subDay()->endOfDay()],
                    'this_week' => $duration = [now()->startOfWeek(), now()->endOfWeek()],
                    'this_month' => $duration = [now()->startOfMonth(), now()->endOfMonth()],
                    'last_month' => $duration = [now()->subMonth()->startOfMonth(), now()->subMonth()->endOfMonth()],
                    'this_year' => $duration = [now()->startOfYear(), now()->endOfYear()],
                    'last_year' => $duration = [now()->subYear()->startOfYear(), now()->subYear()->endOfYear()],
                    'last_week' => $duration = [now()->subWeek()->startOfWeek(), now()->subWeek()->endOfWeek()],
                    default => throw new \Exception("Error Processing Request : '$duration' is not specified in your match code block", 1)
                };
            } catch (\Throwable $th) {
                dd($th->getMessage());
            }
        }

        // using teneray to handle if ($duration == null)
        $requestedOrders = $duration
            ? Order::successful()->whereBetween('created_at', $duration)->get()
            : Order::successful()->get();

        foreach ($requestedOrders as $order)
            $totalFound[$order->id] = $this->findItemInOrder($order);

        return collect($totalFound)->filter(fn($elem) => !empty ($elem)); //filter out empty or null values
    }


    public function getCollectionOfTotalBookSoldToday()
    {
        /**
         * Since we are looking for today's data, we only need to traverse over records created today
         */
        $totalFound = [];

        $today = now()->format('Y-m-d');

        $orderInTimeframe = Order::successful()->whereDate('created_at', $today)->get();
        // dd('today', $orderInTimeframe);
        foreach ($orderInTimeframe as $order) {
            $totalFound[] = $this->findItemInOrder($order);
        } //store collections in array for easy filtration

        return collect($totalFound)->filter(fn($elem) => !empty ($elem)); //filter out empty or null values
    }

    public function getCollectionOfTotalBookSoldYesterday(): Collection
    {
        /**
         * Since we are looking for today's data, we only need to traverse over records created today
         */
        $totalFound = [];

        $yesterday = now()->subDay()->format('Y-m-d');

        $orderInTimeframe = Order::successful()->whereDate('created_at', $yesterday)->get();
        // dd('today', $orderInTimeframe);
        foreach ($orderInTimeframe as $order) {
            $totalFound[] = $this->findItemInOrder($order);
        } //store collections in array for easy filtration

        return collect($totalFound)->filter(fn($elem) => !empty ($elem)); //filter out empty or null values
    }

    /**
     * This might not be needed
     *
     * returns an assoc Array where the key for every client is the [order->date] from the related Order
     */
    public function getUniqueClients()
    {
        $clientIDs = [];
        $clients = [];
        $completedOrders = Order::successful()->get();
        // foreach (Order::where('content', '!=', 'n/a')->successful()->get() as $order) $totalFound[]  = $this->findItemInOrder($order);   //store collections in array for easy filtration
        foreach ($completedOrders as $order) {
            if ($this->findItemInOrder($order)?->id) {
                $clientIDs["{$order->created_at}"] = $order->client_id;
            }
        }

        // we want to sort the clients such that when handling duplicates(same client buying same product multiple times)
        // the latest clients override former ones. This way one can see the latest activity of that client

        //  sort client ids according to the latest (using their keys)
        krsort($clientIDs);

        // get unique clients to avoid duplicates
        $uniqueClientIDs = array_unique($clientIDs);

        foreach ($uniqueClientIDs as $key => $clientID) {
            $clients[$key] = Client::find($clientID);
        }

        return $clients;
    }

    public function getCollectionOfTotalBookSoldLastWeek()
    {
        /**
         * Since we are looking for last week's data, we only need to traverse over records created last week
         * Note this should only happen every monday, so that we can accurately get the cashout for each instructor.
         * Payment cronjob runs daily, but only executes on monday, as its the best time to view the past week.
         */
        $startOfPreviousWeek = now()->subWeek()->startOfWeek();
        $endOfPreviousWeek = now()->subWeek()->endOfWeek();

        return $lastWeekOrders = $this->getOrdersBetween($startOfPreviousWeek, $endOfPreviousWeek);
    }

    public function computeCashoutForPreviousWeek()
    {
        if (strtolower(now()->format('l')) !== 'monday') {
            return false;
        } //we only compute cashout on mondays to avoid confusion
    }

    public function computeCashoutForCurrentWeek()
    {
        // can be done anytime
    }

    public function getCollectionOfTotalBookSoldThisWeek()
    {
        /**
         * Since we are looking for this week's data, we only need to traverse over records created in the current week
         * Note this can happen anytime in the current week, so that we can accurately get the cashout for each instructor per week
         */
        $startOfCurrentWeek = now()->startOfWeek();
        $endOfCurrentWeek = now()->endOfWeek();

        // dd($startOfCurrentWeek->format('Y-m-d'), $endOfCurrentWeek);

        return $thisWeekOrders = $this->getOrdersBetween($startOfCurrentWeek, $endOfCurrentWeek);
    }

    public function getCollectionOfTotalBookSoldLastWeek__fake()
    {
        // going into the future by one day
        if (strtolower(Carbon::tomorrow()->format('l')) !== 'monday') {
            return 'can only compute income on monday' . now()->addDay()->format('l');
        }
        $startOfPreviousWeek = Carbon::tomorrow()->subWeek()->startOfWeek();
        $endOfPreviousWeek = Carbon::tomorrow()->subWeek()->endOfWeek();

        return $lastWeekOrders_fake = $this->getOrdersBetween($startOfPreviousWeek, $endOfPreviousWeek);
    }

    public function getCollectionOfTotalBookSoldLastMonth()
    {
        /**
         * Since we are looking for last month's data, we only need to traverse over records created last month
         */
        $totalFound = [];
        $startOfPreviousMonth = now()->subMonth()->startOfMonth();
        $endOfPreviousMonth = now()->subMonth()->endOfMonth();

        /**
         * @var $lastMonthOrders refers to orders made in the last 30 days or one month ago
         */

        return $lastMonthOrders = $this->getOrdersBetween($startOfPreviousMonth, $endOfPreviousMonth);
    }

    public function getCollectionOfTotalBookSoldThisMonth()
    {
        /**
         * Since we are looking for this month's data, we only need to traverse over records created in current month
         */
        $totalFound = [];
        $startOfCurrentMonth = now()->startOfMonth();
        $endOfCurrentMonth = now()->endOfMonth();

        /**
         * @var $lastMonthOrders refers to orders made in the last 30 days or one month ago
         */

        return $CurrentMonthOrders = $this->getOrdersBetween($startOfCurrentMonth, $endOfCurrentMonth);
    }

    public function getCollectionOfTotalBookSoldLastYear()
    {
        /**
         * Since we are looking for last month's data, we only need to traverse over records created last month
         */
        $totalFound = [];
        $startOfPreviousYear = now()->subYear()->startOfYear();
        $endOfPreviousYear = now()->subYear()->endOfYear();

        /**
         * @var $lastYearOrders refers to orders made in the last 30 days or one Year ago
         */
        return $lastYearOrders = $this->getOrdersBetween($startOfPreviousYear, $endOfPreviousYear);
    }

    public function getCollectionOfTotalBookSoldThisYear()
    {
        /**
         * Since we are looking for last month's data, we only need to traverse over records created last month
         */
        $totalFound = [];
        $startOfCurrentYear = now()->startOfYear();
        $endOfCurrentYear = now()->endOfYear();

        /**
         * @var $lastYearOrders refers to orders made in the last 30 days or one Year ago
         */
        return $CurrentYearOrders = $this->getOrdersBetween($startOfCurrentYear, $endOfCurrentYear);
    }
}
